from lightman_ai.sources.base import BaseSource
from lightman_ai.sources.bleeping_computer import BleepingComputerSource
from lightman_ai.sources.the_hacker_news import TheHackerNewsSource

# Mapping of source names to their corresponding classes
SOURCE_REGISTRY = {"the_hacker_news": TheHackerNewsSource, "bleeping_computer": BleepingComputerSource}

SOURCE_CHOICES = list(SOURCE_REGISTRY.keys())


def get_source_class_from_source_name(source_name: str) -> type[BaseSource]:
    """
    Get the source class for a given source name.

    Args:
        source_name: The name of the source (e.g., "the_hacker_news")

    Returns:
        The source class corresponding to the source name

    Raises:
        ValueError: If the source name is not supported
    """
    if source_name not in SOURCE_REGISTRY:
        supported_sources = ", ".join(SOURCE_CHOICES)
        raise ValueError(f"Unsupported source: {source_name}. Supported sources: {supported_sources}")

    return SOURCE_REGISTRY[source_name]



from lightman_ai.core.exceptions import BaseLightmanError


class MultipleDateSourcesError(BaseLightmanError):
    """Exception for when more than one date source is provided."""


class NoSourcesError(BaseLightmanError):
    """Exception for when no sources have been specified."""



import logging
from abc import ABC, abstractmethod
from typing import ClassVar, Never, override

from lightman_ai.ai.base.prompts import MERGE_ARTICLES_FROM_DIFFERENT_SOURCES
from lightman_ai.article.models import SelectedArticlesList
from pydantic_ai import Agent
from pydantic_ai.models.google import GoogleModel
from pydantic_ai.models.openai import OpenAIChatModel


class BaseAgent(ABC):
    _AGENT_CLASS: type[OpenAIChatModel] | type[GoogleModel]
    _DEFAULT_MODEL_NAME: str
    _AGENT_NAME: ClassVar[str]

    def __init__(
        self,
        multiple_sources: bool,
        system_prompt: str,
        model: str | None = None,
        logger: logging.Logger | None = None,
    ) -> None:
        selected_model = model or self._DEFAULT_MODEL_NAME
        agent_model = self._AGENT_CLASS(selected_model)
        self.agent: Agent[Never, SelectedArticlesList] = Agent(
            model=agent_model,
            output_type=SelectedArticlesList,
            system_prompt=system_prompt,
            instructions=MERGE_ARTICLES_FROM_DIFFERENT_SOURCES if multiple_sources else None,
        )
        self.logger = logger or logging.getLogger("lightman")
        self.logger.info("Selected %s's %s model", self, selected_model)

    @override
    def __str__(self) -> str:
        return self._AGENT_NAME

    @abstractmethod
    async def run_prompt(self, prompt: str) -> SelectedArticlesList: ...



import logging
from datetime import UTC, datetime
from typing import override
from xml.etree import ElementTree

import httpx
import stamina
from httpx import AsyncClient
from lightman_ai.article.models import Article, ArticlesList
from lightman_ai.sources.base import BaseSource
from lightman_ai.sources.exceptions import IncompleteArticleFromSourceError, MalformedSourceResponseError
from pydantic import ValidationError

logger = logging.getLogger("lightman")

_RETRY_ON = httpx.TransportError
_ATTEMPTS = 5
_TIMEOUT = 5

BLEEPING_COMPUTER_URL = "https://news.google.com/rss/search?q=site:bleepingcomputer.com&hl=en-US&gl=US&ceid=US:en"


class BleepingComputerSource(BaseSource):
    @override
    async def get_articles(self, date: datetime | None = None) -> ArticlesList:
        """Return the articles that are present in BleepingComputer feed."""
        logger.info("Downloading articles from %s", BLEEPING_COMPUTER_URL)
        feed = await self.get_feed()
        articles = self._xml_to_list_of_articles(feed)
        logger.info("Articles properly downloaded and parsed.")
        if date:
            return ArticlesList.get_articles_from_date_onwards(articles=articles, start_date=date)
        else:
            return ArticlesList(articles=articles)

    async def get_feed(self) -> str:
        """Retrieve the BleepingComputer RSS Feed."""
        async with AsyncClient() as http_client:
            for attempt in stamina.retry_context(
                on=_RETRY_ON,
                attempts=_ATTEMPTS,
                timeout=_TIMEOUT,
            ):
                with attempt:
                    bleeping_computer_feed = await http_client.get(BLEEPING_COMPUTER_URL)
                    bleeping_computer_feed.raise_for_status()
        return bleeping_computer_feed.text

    def _xml_to_list_of_articles(self, xml: str) -> list[Article]:
        try:
            root = ElementTree.fromstring(xml)
        except ElementTree.ParseError as e:
            raise MalformedSourceResponseError(f"Invalid XML format: {e}") from e

        channel = root.find("channel")
        if channel is None:
            raise MalformedSourceResponseError("No channel element found in RSS feed")

        items = channel.findall("item")
        parsed = []

        for item in items:
            try:
                title = item.findtext("title", default="").strip()
                description = self._clean(item.findtext("description", default="").strip())
                link = item.findtext("link", default="").strip()
                published_at_str = item.findtext("pubDate", default="").strip()

                if not published_at_str:
                    logger.exception("Missing publication date. link: `%s`", link)
                    raise IncompleteArticleFromSourceError()

                published_at = datetime.strptime(published_at_str, "%a, %d %b %Y %H:%M:%S %Z")
                if published_at.tzinfo is None:
                    published_at = published_at.replace(tzinfo=UTC)

                parsed.append(Article(title=title, description=description, link=link, published_at=published_at))

            except (ValidationError, ValueError, IncompleteArticleFromSourceError) as e:
                logger.warning("Failed to parse article. title: `%s`, link: `%s`, error: %s", title, link, str(e))

        return parsed

    @staticmethod
    def _clean(text: str) -> str:
        """Remove CDATA wrappers and non-useful characters."""
        # Remove CDATA wrapper if present
        if text.startswith("<![CDATA[") and text.endswith("]]>"):
            text = text[9:-3]  # Remove <![CDATA[ and ]]>

        # Clean up formatting
        return text.replace("\\n", "").replace("       ", "").strip()



import asyncio
import logging
from datetime import datetime

from lightman_ai.ai.base.agent import BaseAgent
from lightman_ai.ai.utils import get_agent_class_from_agent_name
from lightman_ai.article.models import ArticlesList, PrimarySelectedArticle, SelectedArticlesList
from lightman_ai.exceptions import NoSourcesError
from lightman_ai.integrations.service_desk.integration import (
    ServiceDeskIntegration,
)
from lightman_ai.sources.utils import get_source_class_from_source_name

logger = logging.getLogger("lightman")
logger.addHandler(logging.NullHandler())


async def _get_articles_from_source(source_name: str, start_date: datetime | None = None) -> ArticlesList:
    source_class = get_source_class_from_source_name(source_name)
    source_instance = source_class()
    logger.info("Retrieving articles from %s", source_class)
    return await source_instance.get_articles(start_date)


async def _classify_articles(articles: ArticlesList, agent: BaseAgent) -> SelectedArticlesList:
    return await agent.run_prompt(prompt=str(articles))


async def _create_service_desk_issues(
    selected_articles: list[PrimarySelectedArticle],
    service_desk_client: ServiceDeskIntegration,
    service_desk_project_key: str,
    service_desk_request_id_type: str,
) -> None:
    async def schedule_task(article: PrimarySelectedArticle) -> None:
        try:
            if article.related_articles:
                related_articles_raw = "\n".join(
                    [f"{related_article.title}: {related_article.link}" for related_article in article.related_articles]
                )
                related_articles = f"*Related Articles:*\n{related_articles_raw}\n\n"
            else:
                related_articles = ""

            description = f"*Why is relevant:*\n{article.why_is_relevant}\n\n*Source:* {article.link}\n\n{related_articles}*Score:* {article.relevance_score}/10"
            await service_desk_client.create_request_of_type(
                project_key=service_desk_project_key,
                summary=article.title,
                description=description,
                request_id_type=service_desk_request_id_type,
            )
            logger.info("Created issue for article %s", article.link)
        except Exception:
            logger.exception("Could not create ServiceDesk issue: %s, %s", article.title, article.link)
            raise

    tasks = []
    for article in selected_articles:
        tasks.append(schedule_task(article))
    results = await asyncio.gather(*tasks, return_exceptions=True)
    errors = [result for result in results if isinstance(result, Exception)]
    if errors:
        raise ExceptionGroup("Could not create all ServiceDesk issues", errors)


async def lightman(
    agent: str,
    prompt: str,
    score_threshold: int,
    sources: list[str] | None = None,
    service_desk_project_key: str | None = None,
    service_desk_request_id_type: str | None = None,
    dry_run: bool = False,
    model: str | None = None,
    start_date: datetime | None = None,
) -> list[PrimarySelectedArticle]:
    if not sources:
        raise NoSourcesError

    articles = ArticlesList()
    for source in sources:
        articles += await _get_articles_from_source(source, start_date)

    multiple_sources = len(sources) > 1
    agent_class = get_agent_class_from_agent_name(agent)
    agent_instance = agent_class(multiple_sources, prompt, model, logger=logger)

    classified_articles = await _classify_articles(
        articles=articles,
        agent=agent_instance,
    )

    selected_articles: list[PrimarySelectedArticle] = classified_articles.get_articles_with_score_gte_threshold(
        score_threshold
    )

    if not dry_run:
        if not service_desk_project_key or not service_desk_request_id_type:
            raise ValueError("Missing Service Desk's project key or request id type")

        service_desk_client = ServiceDeskIntegration.from_env()
        await _create_service_desk_issues(
            selected_articles=selected_articles,
            service_desk_client=service_desk_client,
            service_desk_project_key=service_desk_project_key,
            service_desk_request_id_type=service_desk_request_id_type,
        )

    return selected_articles



import asyncio
import json
import logging
import os
from datetime import date
from importlib import metadata

import click
from dotenv import load_dotenv
from lightman_ai.ai.utils import AGENT_CHOICES
from lightman_ai.constants import (
    DEFAULT_AGENT,
    DEFAULT_CONFIG_FILE,
    DEFAULT_CONFIG_SECTION,
    DEFAULT_ENV_FILE,
    DEFAULT_LOG_LEVEL,
    DEFAULT_SCORE,
    DEFAULT_TIME_ZONE,
    VERBOSE_LOG_LEVEL,
)
from lightman_ai.core.config import FileConfig, FinalConfig, PromptConfig
from lightman_ai.core.exceptions import ConfigNotFoundError, InvalidConfigError, PromptNotFoundError
from lightman_ai.core.sentry import configure_sentry
from lightman_ai.exceptions import MultipleDateSourcesError
from lightman_ai.main import lightman
from lightman_ai.sources.utils import SOURCE_CHOICES
from lightman_ai.utils import get_start_date

logger = logging.getLogger("lightman")
logging.basicConfig(
    format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


def get_version() -> str:
    """Read version from VERSION file."""
    return metadata.version("lightman-ai")


@click.group()
@click.version_option(version=get_version(), prog_name="lightman-ai")
def entry_point() -> None:
    pass


@entry_point.command()
@click.option("--agent", type=click.Choice(AGENT_CHOICES), help=("Which agent to use"))
@click.option(
    "--prompt-file",
    type=str,
    default=DEFAULT_CONFIG_FILE,
    help=(f"Location of the config file containing the prompts. Defaults to `{DEFAULT_CONFIG_FILE}`."),
)
@click.option("--prompt", type=str, help=("Which prompt to use"))
@click.option("--model", type=str, default=None, help=("Which model to use. Must be set in conjunction with --agent."))
@click.option(
    "--source",
    "sources",
    type=click.Choice(SOURCE_CHOICES),
    multiple=True,
    help=("Which news sources to use (can specify multiple)"),
)
@click.option(
    "--score",
    type=int,
    help=("The minimum score relevance that an article needs to have to be considered relevant"),
    default=None,
)
@click.option(
    "--config-file",
    type=str,
    default=DEFAULT_CONFIG_FILE,
    help=(f"The config file path. Defaults to `{DEFAULT_CONFIG_FILE}`."),
)
@click.option(
    "--config",
    type=str,
    default=DEFAULT_CONFIG_SECTION,
    help=(f"The config settings to use. Defaults to `{DEFAULT_CONFIG_SECTION}`."),
)
@click.option(
    "--env-file",
    type=str,
    default=None,
    help=(f"Path to the environment file. Defaults to `{DEFAULT_ENV_FILE}`."),
)
@click.option(
    "--dry-run",
    is_flag=True,
    help=(
        "When set, runs the script without publishing the results to the integrated services, just shows them in stdout."
    ),
)
@click.option("--start-date", type=click.DateTime(formats=["%Y-%m-%d"]), help="Start date to retrieve articles")
@click.option("--today", is_flag=True, help="Retrieve articles from today.")
@click.option("--yesterday", is_flag=True, help="Retrieve articles from yesterday.")
@click.option("--json", "output_json", is_flag=True, help="Output results in JSON format.")
@click.option("-v", is_flag=True, help="Be more verbose on output.")
def run(
    agent: str,
    prompt: str,
    prompt_file: str,
    model: str | None,
    sources: tuple[str, ...],
    score: int | None,
    config_file: str,
    config: str,
    env_file: str | None,
    dry_run: bool,
    start_date: date | None,
    today: bool,
    yesterday: bool,
    output_json: bool,
    v: bool,
) -> int:
    """
    Entrypoint of the application.

    Holds no logic. It loads the configuration, calls the main method and returns 0 when succesful .
    """
    load_dotenv(env_file or DEFAULT_ENV_FILE)

    if v:
        logger.setLevel(VERBOSE_LOG_LEVEL)
    else:
        try:
            env_log_level = os.getenv("LOG_LEVEL")
            log_level = env_log_level.upper() if env_log_level else DEFAULT_LOG_LEVEL
            logger.setLevel(log_level)
        except ValueError:
            logger.setLevel(DEFAULT_LOG_LEVEL)
            logger.warning("Invalid logging level. Using default value.")

    configure_sentry(logger.level)

    try:
        start_datetime = get_start_date(os.getenv("TIME_ZONE", DEFAULT_TIME_ZONE), yesterday, today, start_date)
    except MultipleDateSourcesError as e:
        raise click.UsageError(e.args[0]) from e

    try:
        prompt_config = PromptConfig.get_config_from_file(path=prompt_file)
        config_from_file = FileConfig.get_config_from_file(config_section=config, path=config_file)
        final_config = FinalConfig.init_from_dict(
            data={
                "agent": agent or config_from_file.agent or DEFAULT_AGENT,
                "prompt": prompt or config_from_file.prompt,
                "score_threshold": score or config_from_file.score_threshold or DEFAULT_SCORE,
                "model": model or config_from_file.model,
                "sources": list(sources) or config_from_file.sources or SOURCE_CHOICES,
            }
        )
        prompt_text = prompt_config.get_prompt(final_config.prompt)
    except (InvalidConfigError, PromptNotFoundError, ConfigNotFoundError) as err:
        raise click.BadParameter(err.args[0]) from None

    relevant_articles = asyncio.run(
        lightman(
            agent=final_config.agent,
            prompt=prompt_text,
            score_threshold=final_config.score_threshold,
            sources=final_config.sources,
            dry_run=dry_run,
            service_desk_project_key=config_from_file.service_desk_project_key,
            service_desk_request_id_type=config_from_file.service_desk_request_id_type,
            model=final_config.model,
            start_date=start_datetime,
        )
    )

    if output_json:
        if relevant_articles:
            articles_data = [article.model_dump() for article in relevant_articles]
            click.echo(json.dumps({"articles": articles_data}, indent=2, default=str))
        else:
            click.echo(json.dumps({"articles": []}))
    else:
        relevant_articles_metadata = []
        for article in relevant_articles:
            related_articles_raw = "\n".join(
                [f" - {related.title} ({related.link})" for related in article.related_articles]
            )
            related_articles = f"\n Related Articles:\n {related_articles_raw}" if related_articles_raw else ""
            relevant_articles_metadata.append(f"{article.title} ({article.link}){related_articles}")

        if relevant_articles_metadata:
            articles = f"Found these articles:\n* {'\n\n* '.join(relevant_articles_metadata)} "
            click.echo(click.style(articles))
        else:
            click.echo(click.style("No relevant articles found."))

    return 0



MERGE_ARTICLES_FROM_DIFFERENT_SOURCES = """
I'm sending you a set of articles, potentially from multiple sources.
A source is the website where an article originates.
Some articles may be duplicates across different sources.
For these duplicated articles, please select one to be the primary article and list the others as its related articles.
"""
